import argparse
import logging
import logging.config

from typing import Dict

import yaml


# =====
def init() -> Dict:
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", default="kvmd.yaml", metavar="<path>")
    options = parser.parse_args()

    with open(options.config) as config_file:
        config = yaml.load(config_file)

    logging.captureWarnings(True)
    logging.config.dictConfig(config["logging"])

    return config["kvmd"]



import os
import signal
import asyncio
import time

from typing import List
from typing import Set
from typing import Callable
from typing import Optional

import aiohttp.web

from .atx import Atx
from .streamer import Streamer
from .ps2 import Ps2Keyboard

from .logging import get_logger


# =====
def _system_task(method: Callable) -> Callable:
    async def wrap(self: "Server") -> None:
        try:
            await method(self)
        except asyncio.CancelledError:
            pass
        except Exception:
            get_logger().exception("Unhandled exception, killing myself ...")
            os.kill(os.getpid(), signal.SIGTERM)
    return wrap


class Server:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        atx: Atx,
        streamer: Streamer,
        keyboard: Ps2Keyboard,
        heartbeat: float,
        atx_leds_poll: float,
        video_shutdown_delay: float,
        loop: asyncio.AbstractEventLoop,
    ) -> None:

        self.__atx = atx
        self.__streamer = streamer
        self.__heartbeat = heartbeat
        self.__keyboard = keyboard
        self.__video_shutdown_delay = video_shutdown_delay
        self.__atx_leds_poll = atx_leds_poll
        self.__loop = loop

        self.__sockets: Set[aiohttp.web.WebSocketResponse] = set()
        self.__sockets_lock = asyncio.Lock()

        self.__system_tasks: List[asyncio.Task] = []

    def run(self, host: str, port: int) -> None:
        self.__keyboard.start()

        app = aiohttp.web.Application(loop=self.__loop)
        app.router.add_get("/", self.__root_handler)
        app.router.add_get("/ws", self.__ws_handler)
        app.on_shutdown.append(self.__on_shutdown)
        app.on_cleanup.append(self.__on_cleanup)

        self.__system_tasks.extend([
            self.__loop.create_task(self.__monitor_keyboard()),
            self.__loop.create_task(self.__stream_controller()),
            self.__loop.create_task(self.__poll_dead_sockets()),
            self.__loop.create_task(self.__poll_atx_leds()),
        ])

        aiohttp.web.run_app(
            app=app,
            host=host,
            port=port,
            print=(lambda text: [get_logger().info(line.strip()) for line in text.strip().splitlines()]),  # type: ignore
        )

    async def __root_handler(self, _: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(text="OK")

    async def __ws_handler(self, request: aiohttp.web.Request) -> aiohttp.web.WebSocketResponse:
        ws = aiohttp.web.WebSocketResponse(heartbeat=self.__heartbeat)
        await ws.prepare(request)
        await self.__register_socket(ws)
        async for msg in ws:
            if msg.type == aiohttp.web.WSMsgType.TEXT:
                retval = await self.__execute_command(msg.data)
                if retval:
                    await ws.send_str(retval)
            else:
                break
        return ws

    async def __on_shutdown(self, _: aiohttp.web.Application) -> None:
        logger = get_logger(0)

        logger.info("Cancelling system tasks ...")
        for task in self.__system_tasks:
            task.cancel()
        await asyncio.gather(*self.__system_tasks)

        logger.info("Disconnecting clients ...")
        for ws in list(self.__sockets):
            await self.__remove_socket(ws)

    async def __on_cleanup(self, _: aiohttp.web.Application) -> None:
        if self.__keyboard.is_alive():
            self.__keyboard.stop()
        if self.__streamer.is_running():
            await self.__streamer.stop()

    @_system_task
    async def __monitor_keyboard(self) -> None:
        while self.__keyboard.is_alive():
            await asyncio.sleep(0.1)
        raise RuntimeError("Keyboard dead")

    @_system_task
    async def __stream_controller(self) -> None:
        prev = 0
        shutdown_at = 0.0
        while True:
            cur = len(self.__sockets)
            if prev == 0 and cur > 0:
                if not self.__streamer.is_running():
                    await self.__streamer.start()
            elif prev > 0 and cur == 0:
                shutdown_at = time.time() + self.__video_shutdown_delay
            elif prev == 0 and cur == 0 and time.time() > shutdown_at:
                if self.__streamer.is_running():
                    await self.__streamer.stop()
            prev = cur
            await asyncio.sleep(0.1)

    @_system_task
    async def __poll_dead_sockets(self) -> None:
        while True:
            for ws in list(self.__sockets):
                if ws.closed or not ws._req.transport:  # pylint: disable=protected-access
                    await self.__remove_socket(ws)
            await asyncio.sleep(0.1)

    @_system_task
    async def __poll_atx_leds(self) -> None:
        while True:
            if self.__sockets:
                await self.__broadcast("EVENT atx_leds %d %d" % (self.__atx.get_leds()))
            await asyncio.sleep(self.__atx_leds_poll)

    async def __broadcast(self, msg: str) -> None:
        await asyncio.gather(*[
            ws.send_str(msg)
            for ws in list(self.__sockets)
            if not ws.closed and ws._req.transport  # pylint: disable=protected-access
        ], return_exceptions=True)

    async def __execute_command(self, command: str) -> Optional[str]:
        (command, args) = (command.strip().split(" ", maxsplit=1) + [""])[:2]
        if command == "CLICK":
            method = {
                "power": self.__atx.click_power,
                "power_long": self.__atx.click_power_long,
                "reset": self.__atx.click_reset,
            }.get(args)
            if method:
                await method()
                return None
        get_logger().warning("Received an incorrect command: %r", command)
        return "ERROR incorrect command"

    async def __register_socket(self, ws: aiohttp.web.WebSocketResponse) -> None:
        async with self.__sockets_lock:
            self.__sockets.add(ws)
            get_logger().info("Registered new client socket: remote=%s; id=%d; active=%d",
                              ws._req.remote, id(ws), len(self.__sockets))  # pylint: disable=protected-access

    async def __remove_socket(self, ws: aiohttp.web.WebSocketResponse) -> None:
        async with self.__sockets_lock:
            try:
                self.__sockets.remove(ws)
                get_logger().info("Removed client socket: remote=%s; id=%d; active=%d",
                                  ws._req.remote, id(ws), len(self.__sockets))  # pylint: disable=protected-access
                await ws.close()
            except Exception:
                pass



import multiprocessing
import multiprocessing.queues
import queue
import time

from .logging import get_logger

from . import gpio


# =====
class Ps2Keyboard(multiprocessing.Process):
    def __init__(self, clock: int, data: int, pulse: float) -> None:
        super().__init__(daemon=True)

        self.__clock = gpio.set_output(clock, initial=True)
        self.__data = gpio.set_output(data, initial=True)
        self.__pulse = pulse

        self.__queue: multiprocessing.queues.Queue = multiprocessing.Queue()
        self.__event = multiprocessing.Event()

    def start(self) -> None:
        get_logger().info("Starting keyboard daemon ...")
        super().start()

    def stop(self) -> None:
        get_logger().info("Stopping keyboard daemon ...")
        self.__event.set()
        self.join()

    def send_byte(self, code: int) -> None:
        self.__queue.put(code)

    def run(self) -> None:
        with gpio.bcm():
            try:
                while not self.__event.is_set():
                    try:
                        code = self.__queue.get(timeout=0.1)
                    except queue.Empty:
                        pass
                    else:
                        self.__send_byte(code)
            except Exception:
                get_logger().exception("Unhandled exception")
                raise

    def __send_byte(self, code: int) -> None:
        code_bits = list(map(bool, bin(code)[2:].zfill(8)))
        code_bits.reverse()
        message = [False] + code_bits + [(not sum(code_bits) % 2), True]
        for bit in message:
            self.__send_bit(bit)

    def __send_bit(self, bit: bool) -> None:
        gpio.write(self.__clock, True)
        gpio.write(self.__data, bool(bit))
        time.sleep(self.__pulse)
        gpio.write(self.__clock, False)
        time.sleep(self.__pulse)
        gpio.write(self.__clock, True)



import asyncio
import asyncio.subprocess

from typing import Optional

from .logging import get_logger

from . import gpio


# =====
class Streamer:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        cap_power: int,
        conv_power: int,
        sync_delay: float,
        cmd: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:

        self.__cap_power = gpio.set_output(cap_power)
        self.__conv_power = (gpio.set_output(conv_power) if conv_power > 0 else conv_power)
        self.__sync_delay = sync_delay
        self.__cmd = cmd

        self.__loop = loop

        self.__proc_task: Optional[asyncio.Task] = None

    async def start(self) -> None:
        assert not self.__proc_task
        get_logger().info("Starting streamer ...")
        await self.__set_hw_enabled(True)
        self.__proc_task = self.__loop.create_task(self.__process())

    async def stop(self) -> None:
        assert self.__proc_task
        get_logger().info("Stopping streamer ...")
        self.__proc_task.cancel()
        await asyncio.gather(self.__proc_task, return_exceptions=True)
        await self.__set_hw_enabled(False)
        self.__proc_task = None

    def is_running(self) -> bool:
        return bool(self.__proc_task)

    async def __set_hw_enabled(self, enabled: bool) -> None:
        # XXX: This sequence is very important to enable converter and cap board
        gpio.write(self.__cap_power, enabled)
        if self.__conv_power > 0:
            if enabled:
                await asyncio.sleep(self.__sync_delay)
            gpio.write(self.__conv_power, enabled)
        await asyncio.sleep(self.__sync_delay)

    async def __process(self) -> None:
        logger = get_logger(0)

        proc: Optional[asyncio.subprocess.Process] = None  # pylint: disable=no-member
        while True:
            try:
                proc = await asyncio.create_subprocess_shell(
                    self.__cmd,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.STDOUT,
                )
                logger.info("Started streamer pid=%d: %s", proc.pid, self.__cmd)

                empty = 0
                while proc.returncode is None:
                    line = (await proc.stdout.readline()).decode(errors="ignore").strip()
                    if line:
                        logger.info("streamer: %s", line)
                        empty = 0
                    else:
                        empty += 1
                        if empty == 100:  # asyncio bug
                            break

                await self.__kill(proc)
                raise RuntimeError("WTF")

            except asyncio.CancelledError:
                break
            except Exception as err:
                if proc:
                    logger.exception("Unexpected finished streamer pid=%d with retcode=%d", proc.pid, proc.returncode)
                else:
                    logger.exception("Can't start streamer: %s", err)
                await asyncio.sleep(1)

        if proc:
            await self.__kill(proc)

    async def __kill(self, proc: asyncio.subprocess.Process) -> None:  # pylint: disable=no-member
        try:
            proc.terminate()
            await asyncio.sleep(1)
            if proc.returncode is None:
                proc.kill()
            await proc.wait()
        except Exception:
            pass



import contextlib

from typing import Generator

from RPi import GPIO

from .logging import get_logger


# =====
@contextlib.contextmanager
def bcm() -> Generator[None, None, None]:
    logger = get_logger(2)
    GPIO.setmode(GPIO.BCM)
    logger.info("Configured GPIO mode as BCM")
    try:
        yield
    finally:
        GPIO.cleanup()
        logger.info("GPIO cleaned")


def set_output(pin: int, initial: bool=False) -> int:
    GPIO.setup(pin, GPIO.OUT, initial=initial)
    GPIO.output(pin, False)
    return pin


def set_input(pin: int) -> int:
    GPIO.setup(pin, GPIO.IN)
    return pin


def read(pin: int) -> bool:
    return bool(GPIO.input(pin))


def write(pin: int, flag: bool) -> None:
    GPIO.output(pin, flag)



import asyncio

from typing import Tuple

from .logging import get_logger

from . import gpio


# =====
class Atx:
    def __init__(
        self,
        power_led: int,
        hdd_led: int,
        power_switch: int,
        reset_switch: int,
        click_delay: float,
        long_click_delay: float,
    ) -> None:

        self.__power_led = gpio.set_input(power_led)
        self.__hdd_led = gpio.set_input(hdd_led)

        self.__power_switch = gpio.set_output(power_switch)
        self.__reset_switch = gpio.set_output(reset_switch)
        self.__click_delay = click_delay
        self.__long_click_delay = long_click_delay

        self.__lock = asyncio.Lock()

    def get_leds(self) -> Tuple[bool, bool]:
        return (
            not gpio.read(self.__power_led),
            not gpio.read(self.__hdd_led),
        )

    async def click_power(self) -> None:
        if (await self.__click(self.__power_switch, self.__click_delay)):
            get_logger().info("Clicked power")

    async def click_power_long(self) -> None:
        if (await self.__click(self.__power_switch, self.__long_click_delay)):
            get_logger().info("Clicked power (long press)")

    async def click_reset(self) -> None:
        if (await self.__click(self.__reset_switch, self.__click_delay)):
            get_logger().info("Clicked reset")

    async def __click(self, pin: int, delay: float) -> bool:
        if not self.__lock.locked():
            async with self.__lock:
                for flag in (True, False):
                    gpio.write(pin, flag)
                    await asyncio.sleep(delay)
                    return True
        return False



import asyncio

from .application import init
from .logging import get_logger

from .atx import Atx
from .streamer import Streamer
from .ps2 import Ps2Keyboard
from .server import Server

from . import gpio


# =====
def main() -> None:
    config = init()
    with gpio.bcm():
        loop = asyncio.get_event_loop()

        atx = Atx(
            power_led=int(config["atx"]["leds"]["pinout"]["power"]),
            hdd_led=int(config["atx"]["leds"]["pinout"]["hdd"]),
            power_switch=int(config["atx"]["switches"]["pinout"]["power"]),
            reset_switch=int(config["atx"]["switches"]["pinout"]["reset"]),
            click_delay=float(config["atx"]["switches"]["click_delay"]),
            long_click_delay=float(config["atx"]["switches"]["long_click_delay"]),
        )

        streamer = Streamer(
            cap_power=int(config["video"]["pinout"]["cap"]),
            conv_power=int(config["video"]["pinout"]["conv"]),
            sync_delay=float(config["video"]["sync_delay"]),
            cmd=str(config["video"]["cmd"]),
            loop=loop,
        )

        keyboard = Ps2Keyboard(
            clock=int(config["keyboard"]["pinout"]["clock"]),
            data=int(config["keyboard"]["pinout"]["data"]),
            pulse=float(config["keyboard"]["pulse"]),
        )

        Server(
            atx=atx,
            streamer=streamer,
            keyboard=keyboard,
            heartbeat=float(config["server"]["heartbeat"]),
            atx_leds_poll=float(config["atx"]["leds"]["poll"]),
            video_shutdown_delay=float(config["video"]["shutdown_delay"]),
            loop=loop,
        ).run(
            host=str(config["server"]["host"]),
            port=int(config["server"]["port"]),
        )

    get_logger().info("Bye-bye")



import sys
import logging


# =====
def get_logger(depth: int=1) -> logging.Logger:
    frame = sys._getframe(1)  # pylint: disable=protected-access
    frames = []
    while frame:
        frames.append(frame)
        frame = frame.f_back
        if len(frames) - 1 >= depth:
            break
    name = frames[depth].f_globals["__name__"]
    return logging.getLogger(name)
